class AssetReference:
    def __init__(self, *paths):
        self.path = "/".join(paths)
        path_parts = self.path.split("/")
        self.id = path_parts[-1]
        self.type = path_parts[-2]
        if len(path_parts) > 2:
            self.parent = AssetReference("/".join(path_parts[:-2]))
        else:
            self.parent = None


class DBEvent:
    def __init__(self, event):
        value = event["value"] or event["oldValue"]
        self.asset = AssetReference(self._trim_document_path(value["name"]))
        self.data = self._raise_values(value["fields"])

    def _trim_document_path(self, path):
        return "/".join(path.split("/")[5:])

    def _raise_value(self, value_obj):
        value_type = next(iter(value_obj))
        if value_type == "mapValue":
            value = self._raise_values(value_obj[value_type])
        elif value_type == "arrayValue":
            value = [self._raise_value(v) for v in value_obj[value_type]["values"]]
        elif value_type == "referenceValue":
            value = AssetReference(self._trim_document_path(value_obj[value_type]))
        else:
            value = value_obj[value_type]
        return value

    def _raise_values(self, fields):
        values = {}
        for key in fields:
            values[key] = self._raise_value(fields[key])
        return values


class ChangeRef:
    def __init__(self, state: AssetReference, uid: str, chid: str):
        self.sid = state.id
        self.uid = uid
        self.chid = chid
        self.name = "change/{}/{}/{}".format(self.sid, self.uid, self.chid)


class PlanRef:
    def __init__(self, state: AssetReference, uid: str, chid: str):
        self.sid = state.id
        self.uid = uid
        self.chid = chid
        self.name = "tf/plan/{}/{}/{}".format(self.sid, self.uid, self.chid)


class ApplyRef:
    def __init__(self, state: AssetReference):
        self.sid = state.id
        self.name = "tf/apply/{}".format(self.sid)


class Repo:
    def __init__(self, configuration: AssetReference):
        self.cid = configuration.id
        self.wid = configuration.parent.id
        self.name = f"{self.wid}-{self.cid}"




import logging
import os
import json
import requests

SERVICE_CATALOG = json.loads(os.getenv("SERVICE_CATALOG", "{}"))


def get_service_account_metadata(field, params=None):
    url = (
        "http://metadata.google.internal"
        f"/computeMetadata/v1/instance/service-accounts/default/{field}"
    )
    return requests.get(
        url=url, headers={"Metadata-Flavor": "Google"}, params=params
    ).content.decode()


def get_service_url(service_name, service_path=None):
    url = SERVICE_CATALOG.get(service_name)
    if not url:
        raise Exception(f"Service '{service_name}' not found in catalog.")
    if service_path:
        url += f"/{service_path}"
    return url


def call(service_name, service_path=None, **kwargs):
    url = get_service_url(service_name, service_path)
    headers = None
    if not os.getenv("LOCAL"):
        token = get_service_account_metadata("identity", {"audience": url})
        headers = {"Authorization": "Bearer " + token}
    r = requests.post(url, headers=headers, json=kwargs)
    if r.status_code >= 500:
        r.raise_for_status()
    return r



import logging
import json
from base64 import b64decode
from functools import wraps
from flask import jsonify, Request, request, g
from requests.exceptions import HTTPError
from ..service import call
from ..models import AssetReference


class User:
    def __init__(self, encoded_user_info):
        user_info = json.loads(b64decode(encoded_user_info + "==="))
        self.id = user_info["user_id"]
        self.path = "users/" + self.id
        self.email = user_info["email"]
        self.email_verified = user_info["email_verified"]


def response(message, **data):
    return jsonify(message=message, **data)


def get_user():
    encoded_user_info = request.headers.get("X-Endpoint-Api-Userinfo")
    if encoded_user_info:
        return User(encoded_user_info)
    return None


def authorize(permission=None):
    def wrap(func):
        @wraps(func)
        def wrapper():
            try:
                user = get_user()
                if not user:
                    return response("Missing authentication information."), 401
                if permission:
                    if not "asset" in request.json:
                        return (
                            response("Missing 'asset' parameter."),
                            400,
                        )
                    asset = AssetReference(request.json["asset"])
                    r = call(
                        "db",
                        "authorize",
                        asset=asset.path,
                        uid=user.id,
                        permission=permission,
                    )
                    r.raise_for_status()
                    if not r.json()["granted"]:
                        return response("Access denied."), 403
                    g.asset = asset
                g.user = user
            except Exception as e:
                logging.exception(e)
                return response("An unexpected error occurred in authorization."), 500
            return func()

        return wrapper

    return wrap


def validate(**kwargs):
    def wrap(func):
        @wraps(func)
        def wrapper():
            try:
                for key in kwargs:
                    valid = kwargs[key](request.json.get(key))
                    if not valid:
                        return response(f"Invalid value for '{key}'"), 400
            except Exception as e:
                logging.exception(e)
                return response("An unexpected error occurred in validation."), 500
            return func()

        return wrapper

    return wrap



import json
from base64 import b64encode
from copy import deepcopy


def encode_content(content):
    return b64encode(json.dumps(content, sort_keys=True, indent=0).encode()).decode()


def deepmerge(x, y):
    overlapping_keys = x.keys() & y.keys()
    for k in overlapping_keys:
        deepmerge(x[k], y[k])
    for k in y.keys() - overlapping_keys:
        x[k] = deepcopy(y[k])
    return x


def chunk_list(lst, n):
    return [lst[i : i + n] for i in range(0, len(lst), n)]



from re import match


def valid_id(id_):
    return (
        isinstance(id_, str)
        and 1 <= len(id_) <= 40
        and "/" not in id_
        and id_ not in [".", ".."]
        and not match("__.*__", id_)
    )


def valid_name(name):
    return isinstance(name, str) and 1 <= len(name) <= 50


def valid_repo(repo):
    return isinstance(repo, str) and all(map(valid_id, repo.split("-")))


def valid_asset(asset):
    if isinstance(asset, str):
        parts = list(filter(None, asset.split("/")))
        return len(parts) >= 2 and all(map(valid_id, parts))
    return False


def valid_chid(chid):
    return (
        isinstance(chid, str)
        and 1 <= len(chid) <= 100
        and match("[a-zA-Z0-9-]+$", chid)
    )


def valid_change_ref(ref):
    if isinstance(ref, str):
        parts = ref.split("/")
        return (
            len(parts) == 4
            and parts[0] == "change"
            and valid_id(parts[1])
            and valid_id(parts[2])
            and valid_chid(parts[3])
        )
    return False


def valid_apply_ref(ref):
    if isinstance(ref, str):
        parts = ref.split("/")
        return (
            len(parts) == 3
            and parts[0] == "tf"
            and parts[1] == "apply"
            and valid_id(parts[2])
        )
    return False


def valid_configuration_changes(changes):
    if not isinstance(changes, list):
        return False
    for change in changes:
        if not isinstance(change, dict):
            return False
        if "rid" not in change or "content" not in change:
            return False
        content = change["content"]
        if isinstance(content, dict) and "provisioner" in content.keys():
            return False
    return True
